{****************************************************************************}
{*                                                                          *}
{*                          IMPORTANT NOTICE                                *}
{*                                                                          *}
{*         (c) Copyright 1995 by Softouch Development Inc.                  *}
{*                                                                          *}
{*                        ALL RIGHTS RESERVED                               *}
{*                                                                          *}
{* 01/15/96 Fixed problem in Run method where WriteFats on error did not    *}
{*			set the variable result. Mismatched elses.                      *}
{****************************************************************************}
unit Format;

interface

uses
	SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
	Forms, Dialogs, WinDos;


type
DiskInformation = Record
	BytesPerSector    : Word;
	SectorsPerCluster : Byte;
	ReservedSectors   : Word;
	FATs              : Byte;
	RootDirectories   : Word;
	TotalSectors      : Word;
	MediaDescriptor   : Byte;
	SectorsPerFAT     : Word;
	SectorsPerTrack   : Word;
	Heads             : Word;
	end;

TrackTableInformation = Record
	Track          : Byte;
	Head           : Byte;
	Sector         : Byte;
	SectorSizeCode : Byte;
	end;

BootInformation = Record
	Name : array[1..8] of Char;
	DiskInfo : DiskInformation;
	HiddenSectors : LongInt;
	SecsInLogicalVol : LongInt;
	DriveNumber : Byte;
	Reserved : Byte;
	BootSignature : Byte;
	SerialNumber : LongInt;
	Volume : array[1..11] of Char;
	FatType : array[1..8] of Char;
	end;

Type
BootArray = array[1..512] of Byte;
DiskInfoArray = array[1..4] of DiskInformation;
TrackTableArray = array[1..65520 div Sizeof(TrackTableInformation)] of TrackTableInformation;
TrackTablePtr = ^TrackTableArray;
MaxArray = array[1..65520] of Byte;
MaxArrayPtr = ^MaxArray;
DPT = Array[1..10] of Byte;
DPTPtr = ^DPT;
{ This are used for properties }
TDrive = (DriveA,DriveB);
TFormatCondition = (Always,Never,WhenNeeded);
TSystemLocation = (DriveC,DriveD,DriveE,DriveF,DriveG,DriveH,DriveI,DriveJ,DriveK,DriveL,DriveM,
DriveN,DriveO,DriveP,DriveQ,DriveR,DriveS,DriveT,DriveU,DriveV,DriveW,DriveX,DriveY,DriveZ);
TMediaType = (BestFit,_360K,_1_2M,_720K,_1_4M);
TRetries = 1..25;
TVerify = (AlwaysVerify,NeverVerify,WhenFormattingVerify);

TFormat = class(TComponent)
	private
	{ Published Properties }
	FClearDisk : Boolean;
	FDrive : TDrive;
	FFormatName : String;
	FFormatCondition : TFormatCondition;
	FMakeBootable : Boolean;
	FMediaType : TMediaType;
	FMsgBoot : String;
	FMsgNonBootable : String;
	FMsgRetryBoot : String;
	FRetries : TRetries;
	FSystemLocation : TSystemLocation;
	FVerify : TVerify;
	FVolume : String;
	FOnBootWrite : TNotifyEvent;
	FOnFatWrite : TNotifyEvent;
	FOnRootWrite : TNotifyEvent;
	FOnSystemWrite : TNotifyEvent;
	FOnFormat : TNotifyEvent;
	FOnVerify : TNotifyEvent;
	{ RunTime Properties }
	FActive : Boolean;
	FActualMediaType : TMediaType;
	FDriveType : TMediaType;
	FErrNoMessage : String;
	FErrNo : Integer;
	FErrorMessage : String;
	FCurrentHead : Byte;
	FCurrentTrack : Byte;
	FHeads : Byte;
	FTracks : Byte;
	FSectors : Byte;
	FProgress : Integer;
	{ Private variables }
	BootInfo : BootInformation;
	NewDPTPtr : DPTPtr;
	OrgDPT : DPT;
	OrgDPTPtr : DPTPtr;
	TrackTable : TrackTablePtr;
	function ResetDrive(Drive : Byte) : Integer;
	function GetDriveType(Drive : Byte; var DriveType : Byte) : Integer;
	function SetDiskType(Drive : Byte; DiskTypeCode : Byte) : Integer;
	function SetMediaType(Drive : Byte; Tracks : Byte; Sectors : Byte; var NewDPTPtr : DPTPtr) : Integer;
	function FormatTrack(Drive : Byte; Sectors : Byte; Track : Byte; Head : Byte; SectorSizeCode : Byte;
	TrackTable : TrackTablePtr) : Integer;
	function VerifyTrack(Drive : Byte; Sectors : Byte; Track : Byte; Head : Byte) : Integer;
	function WriteBoot(Drive : Byte; var Buffer : MaxArrayPtr) : Integer;
	function WriteFats(Drive : Byte; var Buffer : MaxArrayPtr) : Integer;
	function WriteRoot(Drive : Byte; Buffer : MaxArrayPtr) : Integer;
	function WriteDisk(Drive : Byte; Head : Byte; Track : Byte; Sector : Byte; NumSecs : Byte; Buffer : Pointer) : Integer;
	function WriteLogicalSector(Drive : Byte; LogicalSector : Integer; Buffer : Pointer) : Integer;
	function ReadDisk(Drive : Byte; Head : Byte; Track : Byte; Sector : Byte; NumSecs : Byte; Buffer : Pointer) : Integer;
	function SerialNumber : LongInt;
	function MakeBootDisk(Drive : Char) : Integer;
	function CopyFile(Source : String; Dest : String) : Integer;
	function IsMSDOS : Boolean;
	function CheckFormatConditions(Drive : Byte; var TempFormatCondition : TFormatCondition) : Integer;
	function CompareMemory(Value1, Value2 : MaxArrayPtr; Size : Integer) : Integer;
	{ Property editor functions }
	procedure SetFormatName(Value : String);
	procedure SetVolume(Value : String);
	procedure SetMsgBoot(Value : String);
	procedure SetMsgNonBootable(Value : String);
	procedure SetMsgRetryBoot(Value : String);
	{ Private declarations }
	protected
	{ Protected declarations }
	public
	Cancel : Boolean;
	constructor Create(AOwner: TComponent); override;
	Destructor Destroy; override;
	property Active : Boolean read FActive;
	property ActualMediaType : TMediaType read FActualMediaType;
	property DriveType : TMediaType read FDriveType;
	property ErrNoMessage : String read FErrnoMessage;
	property ErrNo : Integer read FErrNo;
	property ErrorMessage : String read FErrorMessage;
	property CurrentHead : Byte read FCurrentHead;
	property CurrentTrack : Byte read FCurrentTrack;
	property Heads : Byte read FHeads;
	property Tracks : Byte read FTracks;
	property Sectors : Byte read FSectors;
	property Progress : Integer read FProgress;
	function Run : Integer;
	function GetErrNoMessage(ErrNo : Integer) : String;
	function GetErrorMessage(Error : Integer) : String;
	{ Public declarations }
	published
	property ClearDisk : Boolean read FClearDisk write FClearDisk Default True;
	property Drive : TDrive read FDrive write FDrive Default DriveA;
	property FormatName : String read FFormatName write SetFormatName;
	property FormatCondition : TFormatCondition read FFormatCondition write FFormatCondition;
	property MakeBootable : Boolean read FMakeBootable write FMakeBootable Default False;
	property MediaType : TMediaType read FMediaType write FMediaType Default BestFit;
	property MsgBoot : String read FMsgBoot write SetMsgBoot;
	property MsgNonBootable : String read FMsgNonBootable write SetMsgNonBootable;
	property MsgRetryBoot : String read FMsgRetryBoot write SetMsgRetryBoot;
	property Retries : TRetries read FRetries write FRetries Default 5;
	property SystemLocation : TSystemLocation read FSystemLocation write FSystemLocation default DriveC;
	property Verify : TVerify read FVerify write FVerify Default NeverVerify;
	property Volume  : String read FVolume write SetVolume;
	property OnBootWrite : TNotifyEvent read FOnBootWrite write FOnBootWrite;
	property OnFatWrite : TNotifyEvent read FOnFatWrite write FOnFatWrite;
	property OnRootWrite : TNotifyEvent read FOnRootWrite write FOnRootWrite;
	property OnSystemWrite : TNotifyEvent read FOnSystemWrite write FOnSystemWrite;
	property OnFormat : TNotifyEvent read FOnFormat write FOnFormat;
	property OnVerify : TNotifyEvent read FOnVerify write FOnVerify;
	{ Published declarations }
	end;

Const
_IOSYS  : string[13] = '\IO.SYS';
_MSDOS  : string[13] = '\MSDOS.SYS';
_IBMBIO : string[13] = '\IBMBIO.SYS';
_IBMDOS : string[13] = '\IBMDOS.SYS';

BootSector : BootArray =
	($eb,$3c,$90,$20,$20,$20,$20,$20,$20,$20,$20,$00,$00,$00,$00,
	 $00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,
	 $00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$fa,$33,$c0,$8e,$d8,$8e,$c0,$8e,$d0,$bc,$00,$7c,$fb,
	 $cd,$13,$72,$74,$32,$e4,$a0,$10,$7c,$f7,$26,$16,$7c,$03,$06,
	 $0e,$7c,$a3,$f6,$7d,$bb,$00,$05,$e8,$8b,$00,$72,$61,$8b,$fb,
	 $be,$dc,$7d,$b9,$0b,$00,$f3,$a6,$75,$55,$8d,$7f,$20,$be,$e7,
	 $7d,$b9,$0b,$00,$f3,$a6,$75,$48,$b8,$20,$00,$f7,$26,$11,$7c,
	 $f7,$36,$0b,$7c,$03,$06,$f6,$7d,$a3,$f8,$7d,$b9,$03,$00,$bb,
	 $00,$07,$50,$53,$51,$e8,$52,$00,$59,$5b,$58,$72,$25,$40,$03,
	 $1e,$0b,$7c,$e2,$ee,$be,$5b,$7d,$e8,$31,$00,$8a,$2e,$15,$7c,
	 $8a,$16,$24,$7c,$33,$c0,$8b,$1e,$f8,$7d,$ea,$00,$00,$70,$00,
	 $be,$27,$7d,$eb,$05,$be,$86,$7d,$eb,$00,$e8,$11,$00,$be,$b1,
	 $7d,$e8,$0b,$00,$32,$e4,$cd,$16,$cd,$19,$be,$4a,$7d,$eb,$ec,
	 $ac,$0a,$c0,$75,$01,$c3,$b4,$0e,$bb,$07,$00,$cd,$10,$eb,$f1,
	 $33,$d2,$f7,$36,$18,$7c,$fe,$c2,$88,$16,$f5,$7d,$33,$d2,$f7,
	 $36,$1a,$7c,$88,$16,$f4,$7d,$a3,$f2,$7d,$8b,$16,$f2,$7d,$b1,
	 $06,$d2,$e6,$0a,$36,$f5,$7d,$8b,$ca,$86,$e9,$8a,$16,$24,$7c,
	 $8a,$36,$f4,$7d,$b8,$01,$02,$cd,$13,$c3,$45,$72,$72,$6f,$72,
	 $20,$72,$65,$73,$65,$74,$74,$69,$6e,$67,$20,$64,$69,$73,$6b,
	 $20,$63,$6f,$6e,$74,$72,$6f,$6c,$6c,$65,$72,$2e,$0d,$0a,$00,
	 $52,$65,$62,$6f,$6f,$74,$20,$66,$61,$69,$6c,$65,$64,$2e,$0d,
	 $0a,$00,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$0d,$0a,$00,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$0d,$0a,$00,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$0d,$0a,$00,$20,$20,$20,$20,
	 $20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,
	 $20,$20,$20,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,
	 $00,$00);

DiskInfo : DiskInfoArray =
	{ Setup DiskInfo for a 360K}
	((BytesPerSector : 512; SectorsPerCluster : 2; ReservedSectors : 1; Fats : 2; RootDirectories : 112;
	  TotalSectors : 720; MediaDescriptor : $fd; SectorsPerFAT : 2; SectorsPerTrack : 9; Heads : 2),
	{ Setup DiskInfo for a 1.2M}
	 (BytesPerSector : 512; SectorsPerCluster : 1; ReservedSectors : 1; Fats : 2; RootDirectories : 224;
	  TotalSectors : 2400; MediaDescriptor : $f9; SectorsPerFAT : 7; SectorsPerTrack : 15; Heads : 2),
	{ Setup DiskInfo for a 720K}
	 (BytesPerSector : 512; SectorsPerCluster : 2; ReservedSectors : 1; Fats : 2; RootDirectories : 112;
	  TotalSectors : 1440; MediaDescriptor : $f9; SectorsPerFAT : 3; SectorsPerTrack : 9; Heads : 2),
	{ Setup DiskInfo for a 1.4MK}
	 (BytesPerSector : 512; SectorsPerCluster : 1; ReservedSectors : 1; Fats : 2; RootDirectories : 224;
	  TotalSectors : 2880; MediaDescriptor : $f0; SectorsPerFAT : 9; SectorsPerTrack : 18; Heads : 2));

procedure Register;

implementation

procedure Register;
begin
	RegisterComponents('System', [TFormat]);
end;

constructor TFormat.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);

	FActive         := False;
	ClearDisk       := True;
	Drive           := DriveA;
	FormatName      := 'Floppy';
	FormatCondition := WhenNeeded;
	MakeBootable    := False;
	MediaType       := BestFit;
	MsgBoot         := 'Booting from floppy. Please wait...';
	MsgNonBootable  := 'Non-System disk or disk error';
	MsgRetryBoot    := 'Replace and press any key to continue';
	Retries         := 5;
	SystemLocation  := DriveC;
	Verify          := NeverVerify;
	Cancel := False;
	TrackTable := Nil;
	OrgDPTPtr := Nil;
	NewDPTPtr := Nil;
end;

Destructor TFormat.Destroy;
begin
	if OrgDPTPtr <> Nil then
		begin
		OrgDPTPtr^ := OrgDPT;
		OrgDPTPtr := Nil;
		end;

	if TrackTable <> Nil then
		FreeMem(TrackTable,Sizeof(TrackTableInformation)*FSectors);
	inherited Destroy;
end;

procedure TFormat.SetFormatName(Value : String);
begin
	FFormatName := Copy(Value,1,8);
end;

procedure TFormat.SetVolume(Value : String);
begin
	FVolume := Copy(Value,1,11);
end;

procedure TFormat.SetMsgBoot(Value : String);
begin
	FMsgBoot := Copy(Value,1,40);
end;

procedure TFormat.SetMsgNonBootable(Value : String);
begin
	FMsgNonBootable := Copy(Value,1,40);
end;

procedure TFormat.SetMsgRetryBoot(Value : String);
begin
	FMsgRetryBoot := Copy(Value,1,40);
end;

function TFormat.Run : Integer;
var
DiskTypeCode : Byte;
CurTrack : Byte;
CurHead : Byte;
CurSector : Byte;
SectorSizeCode : Byte;
BytesPerSector : Integer;
TempDriveType : Byte;
Regs : TRegisters;
Buffer : MaxArrayPtr;
TempFormatCondition : TFormatCondition;
TempDrive : Byte;
begin
	if FActive = True then
		begin
		Result := 98;
		FErrorMessage := GetErrorMessage(Result);
		exit;
		end;
	FActive := True;

	FProgress := 0;
	FErrNo := 0;
	Cancel := False;
	Result := 0;

	{ Save drive incase user tries to change it while formatting }
	TempDrive := Byte(Ord(FDrive));

	{ Reset the Drive controller }
	FErrNo := ResetDrive(TempDrive);
	if FErrNo <> 0 then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ Reset failed }
		Result := 1;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;

	{ Get Drive type. Duh }
	FErrNo := GetDriveType(TempDrive,TempDriveType);
	if FErrNo <> 0 then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ Unable to get drive type }
		Result := 2;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;
	FDriveType := TMediaType(TempDriveType);

	if FDriveType <= BestFit then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ No such drive }
		Result := 3;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;

	if FDriveType > _1_4M then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ Unknown drive type }
		Result := 4;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;

	if FMediaType = BestFit then
		FActualMediaType := FDriveType
	else
		FActualMediaType := FMediaType;

	TempFormatCondition := FFormatCondition;
	Result := CheckFormatConditions(TempDrive,TempFormatCondition);
	if Result <> 0 then
		begin
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
     	exit;
		end;

	DiskTypeCode := 0;
	case FDriveType of
		_360K:
			if FActualMediaType = _360K then DiskTypeCode := 1;
		_1_2M:
			begin
			if FActualMediaType = _360K then DiskTypeCode := 2;
			if FActualMediaType = _1_2M then DiskTypeCode := 3;
			end;
		_720K:
			if FActualMediaType = _720K then DiskTypeCode := 4;
		_1_4M:
			if FActualMediaType >= _720K then DiskTypeCode := 4;
		end;

	if DiskTypeCode = 0 then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ Incompatible media types }
		Result := 5;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;

	{ Not supported on WIN95 }
	FErrNo := SetDiskType(TempDrive,DiskTypeCode);
	if FErrNo <> 0 then
		begin
		if ErrNo = $01 then
			begin
			{ Must be WIN95 }
			FErrNo := 0;
			end
		else
			begin
			FErrNoMessage := GetErrNoMessage(FErrno);
			{ Unable to set disk type }
			Result := 6;
			FErrorMessage := GetErrorMessage(Result);
			FActive := False;
			exit;
			end;
		end;

	FTracks := DiskInfo[Ord(FActualMediaType)].TotalSectors
		div DiskInfo[Ord(FActualMediaType)].SectorsPerTrack
		div DiskInfo[Ord(FActualMediaType)].Heads;
	FSectors := DiskInfo[Ord(FActualMediaType)].SectorsPerTrack;
	FHeads := DiskInfo[Ord(FActualMediaType)].Heads;

	FErrNo := SetMediaType(TempDrive,FTracks,FSectors,NewDPTPtr);
	if FErrNo <> 0 then
		begin
		FErrNoMessage := GetErrNoMessage(FErrno);
		{ Unable to set media type }
		Result := 7;
		FErrorMessage := GetErrorMessage(Result);
		FActive := False;
		exit;
		end;

	{ Get current Drive Parameter Table }
	GetIntVec($1E,Pointer(OrgDPTPtr));
	OrgDPT     := OrgDPTPtr^;  { Save current value }
	OrgDPTPtr^ := NewDPTPtr^;  { Set New value }

	{ Create TrackTable }
	GetMem(TrackTable,Sizeof(TrackTableInformation)*FSectors);

	{ Sector size code 0 = 128 bytes, 1=256 Bytes, 2=512 Bytes, 3=1024 Bytes }
	BytesPerSector := 128;
	SectorSizeCode := 0;
	while (BytesPerSector <> DiskInfo[Ord(FActualMediaType)].BytesPerSector) and (SectorSizeCode < 4) do
		begin
		BytesPerSector := BytesPerSector * 2;
		SectorSizeCode := SectorSizeCode + 1;
	end;

	{ Do this once }
	for CurSector := 1 to FSectors do
		begin
		TrackTable^[CurSector].Sector := CurSector;
		TrackTable^[CurSector].SectorSizeCode := SectorSizeCode;
		end;

	{ Here's where we Format the tracks }
	for CurTrack := 0 to FTracks-1 do
		begin
		FCurrentTrack := CurTrack;
		for CurHead := 0 to FHeads-1 do
			begin
			FCurrentHead := CurHead;

			FProgress := (100*(CurTrack*FHeads+CurHead+1)) div (FTracks*FHeads);

			if TempFormatCondition = Always then
				begin
				if Assigned(FOnFormat) then FOnFormat(Self);
					Application.ProcessMessages;
				if Cancel = True then
					break;

				FErrNo := FormatTrack(TempDrive,FSectors,CurTrack,CurHead,SectorSizeCode,TrackTable);
				if FErrNo <> 0 then
					begin
					Result := 8;
					break;
					end;
				end;

			if (Verify = AlwaysVerify) or ((Verify = WhenFormattingVerify) and (TempFormatCondition = Always)) then
				begin
				if Assigned(FOnVerify) then FOnVerify(Self);
				Application.ProcessMessages;
				if Cancel = True then
					break;

				FErrNo := VerifyTrack(TempDrive,FSectors,CurTrack,CurHead);
				if FErrNo <> 0 then
					begin
					Result := 9;
					break;
					end;
				end;
			end;

		if Cancel = True then
			break;

		if FErrNo <> 0 then
			break;
		end;

	if (Cancel <> True) and (FErrNo = 0)
	and ((TempFormatCondition = Always)
	or ((TempFormatCondition in [Never,WhenNeeded]) and (FClearDisk = True))) then
		begin
		{ Get space for sector }
		GetMem(Buffer,DiskInfo[Ord(FActualMediaType)].BytesPerSector);

		if Assigned(FOnBootWrite) then FOnBootWrite(Self);
		Application.ProcessMessages;
		if Cancel <> True then
			begin
			FErrNo := WriteBoot(TempDrive,Buffer);
			if FErrNo = 0 then
				begin
				if Assigned(FOnFatWrite) then FOnFatWrite(Self);
				Application.ProcessMessages;
				if Cancel <> True then
					begin
					FErrNo := WriteFats(TempDrive,Buffer);
					if FErrNo = 0 then
					    begin
                                            if Assigned(FOnRootWrite) then FOnRootWrite(Self);
					    Application.ProcessMessages;
					    if Cancel <> True then
					    	begin
					    	FErrNo := WriteRoot(TempDrive,Buffer);
					    	if FErrNo <> 0 then
					    		Result := 12;
					    	end;
					    end
					else
                                            Result := 11;
                                        end;
				end
			else
				Result := 10;
			end;

		{ Free sector }
		FreeMem(Buffer,DiskInfo[Ord(FActualMediaType)].BytesPerSector);

		if (FErrNo = 0) and (Cancel <> True) and (FMakeBootable = True) then
			begin
			if Assigned(FOnSystemWrite) then FOnSystemWrite(Self);
			Application.ProcessMessages;
			if Cancel <> True then
				begin
				FErrNo := MakeBootDisk(Char(Integer('A')+TempDrive));
				if FErrNo <> 0 then
					Result := 13;
				end;
			end;
		end;

	FreeMem(TrackTable,Sizeof(TrackTableInformation)*FSectors);
	TrackTable := Nil;

	if OrgDPTPtr <> Nil then
		begin
		OrgDPTPtr^ := OrgDPT;
		OrgDPTPtr := Nil;
		end;
	if Cancel = True then
		Result := 99;
	FErrorMessage := GetErrorMessage(Result);
	FActive := False;
end;

function TFormat.ResetDrive(Drive : Byte) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,sizeof(Regs),0);
		Regs.ah := 0;
		Regs.dl := Drive;
		Intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
        		break;
		end;
end;

function TFormat.GetDriveType(Drive : Byte; var DriveType : Byte) : Integer;
var Regs : TRegisters;
begin
	FillChar(Regs,Sizeof(Regs),0);
	Regs.ah := $8;
	Regs.dl := Drive;
	Intr($13,Regs);
	Result := Regs.ah;
	if Result = 0 then
		DriveType := Regs.bl
	else
		DriveType := 0;
end;

function TFormat.SetDiskType(Drive : Byte; DiskTypeCode : Byte) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $17;
		Regs.dl := Drive;
		Regs.al := DiskTypeCode;
		Intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			break;
		ResetDrive(Drive);
		end;
end;

function TFormat.SetMediaType(Drive : Byte; Tracks : Byte; Sectors : Byte; var NewDPTPtr : DPTPtr) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $18;
		Regs.dl := Drive;
		Regs.ch := Tracks-1;
		Regs.cl := Sectors;
		Intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			begin;
			NewDPTPtr := Ptr(Regs.es,Regs.di);
			break;
			end;
		ResetDrive(Drive);
		end;
end;


function TFormat.FormatTrack(Drive : Byte; Sectors : Byte; Track : Byte; Head : Byte; SectorSizeCode : Byte;
TrackTable : TrackTablePtr) : Integer;
var
Retry : Integer;
Regs : TRegisters;
CurSector : Byte;
begin
	for CurSector := 1 to Sectors do
		begin
		TrackTable^[CurSector].Track := Track;
		TrackTable^[CurSector].head := Head;
		end;

	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $5;
		Regs.al := Sectors;
		Regs.cl := 1;
		Regs.ch := Track;
		Regs.dh := Head;
		Regs.es := Seg(TrackTable^);
		Regs.bx := Ofs(TrackTable^);
		Regs.dl := Drive;
		if Regs.dl <= 1 then { Never can be to safe }
			intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			break;
		ResetDrive(Drive);
		end;
end;

function TFormat.VerifyTrack(Drive : Byte; Sectors : Byte; Track : Byte; Head : Byte) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $4;
		Regs.al := Sectors;
		Regs.cl := 1;
		Regs.ch := Track;
		Regs.dh := Head;
		Regs.dl := Drive;
		intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			break;
		ResetDrive(Drive);
		end;
end;

function TFormat.WriteBoot(Drive : Byte; var Buffer : MaxArrayPtr) : Integer;
var Temp : array[1..41] of Char;
begin
	{ Move our 512 byte sector into complete sector }
	move(BootSector,Buffer^,Sizeof(BootSector));

	{ Fill Remainder with zeros }
	FillChar(Buffer^[Sizeof(BootSector)+3],DiskInfo[Ord(FActualMediaType)].BytesPerSector-Sizeof(BootSector),0);

	{ But signature at the end }
	Buffer^[DiskInfo[Ord(FActualMediaType)].BytesPerSector-1] := Byte(#$55);
	Buffer^[DiskInfo[Ord(FActualMediaType)].BytesPerSector]   := Byte(#$AA);

	{ Setup BootInfo }
	FillChar(BootInfo,Sizeof(BootInfo),0);
	BootInfo.DiskInfo := DiskInfo[Ord(FActualMediaType)];
	BootInfo.BootSignature := $29;
	BootInfo.SerialNumber := SerialNumber;
	BootInfo.FatType := 'FAT12   ';

	{ Fill in users info }

	{ Put name of format }
	StrPLCopy(PChar(@Temp),FFormatName,Length(FFormatName));
	FillChar(BootInfo.Name,Sizeof(BootInfo.Name),' ');
	Move(Temp,BootInfo.Name,Length(FFormatName));

	{ Put Volume Id }
	FillChar(BootInfo.Volume,Sizeof(BootInfo.Volume),' ');
	if Length(FVolume) > 0 then
		begin
		StrPLCopy(PChar(@Temp),FVolume,Length(FVolume));
		Move(Temp,BootInfo.Volume,Length(FVolume));
		end
	else
		begin
		StrPCopy(PChar(@Temp),'NO NAME');
		Move(Temp,BootInfo.Volume,7);
		end;

	{ Put BootInfo into complete boot sector info }
	Move(BootInfo,Buffer^[4],Sizeof(BootInfo));

	if IsMSDOS = True then
		begin
		StrPLCopy(PChar(@Temp),'IO      SYS',11);
		Move(Temp,Buffer^[477],11);
		StrPLCopy(PChar(@Temp),'MSDOS   SYS',11);
		Move(Temp,Buffer^[488],11);
		end
	else
		begin
		StrPLCopy(PChar(@Temp),'IBMBIO  SYS',11);
		Move(Temp,Buffer^[477],11);
		StrPLCopy(PChar(@Temp),'IBMDOS  SYS',11);
		Move(Temp,Buffer^[488],11);
		end;

	{ Put Boot Message }
	StrPLCopy(PChar(@Temp),FMsgBoot,Length(FMsgBoot));
	Move(Temp,Buffer^[348],Length(FMsgBoot));

	{ Put NonBootable Message }
	StrPLCopy(PChar(@Temp),FMsgNonBootable,Length(FMsgNonBootable));
	Move(Temp,Buffer^[391],Length(FMsgNonBootable));

	{ Put RetryBootMsg }
	StrPLCopy(PChar(@Temp),FMsgRetryBoot,Length(FMsgRetryBoot));
	Move(Temp,Buffer^[434],Length(FMsgRetryBoot));

	{ Write out complete sector }
	Result := WriteDisk(Drive,0,0,1,1,Buffer);
end;

function TFormat.WriteFats(Drive : Byte; var Buffer : MaxArrayPtr) : Integer;
var
Fat : Byte;
SecInFat : Byte;
LogicalSector : Integer;
begin
	with DiskInfo[Ord(FActualMediaType)] do
		begin
		for Fat := 1 to Fats do
			begin
			{ Write as many sectors needed for each fat }
			for SecInFat := 1 to SectorsPerFat do
				begin
				{ Clear Buffer }
				FillChar(Buffer^,BytesPerSector,0);
				if SecInFat = 1 then
					begin
					Buffer^[1] := MediaDescriptor;
					Buffer^[2] := $ff;
					Buffer^[3] := Buffer^[2];
					end;
				LogicalSector := ReservedSectors + ((Fat-1)*SectorsPerFat) + (SecInFat-1);
				Result := WriteLogicalSector(Drive,LogicalSector,Buffer);
				if Result <> 0 then
					exit;
				end;
			end;
		end;
end;

function TFormat.WriteRoot(Drive : Byte; Buffer : MaxArrayPtr) : Integer;
var
StartSector : Integer;
EndSector : Integer;
LogicalSector : Integer;
Temp : array[1..11] of Byte;
FileTime : LongInt;
Offset : Integer;
begin
	with DiskInfo[Ord(FActualMediaType)] do
		begin
		StartSector := ReservedSectors + Fats * SectorsPerFat;
		EndSector := (RootDirectories * 32) div BytesPerSector + StartSector;
		for LogicalSector := StartSector to EndSector do
			begin
			FillChar(Buffer^,BytesPerSector,0);
			if (LogicalSector = StartSector) and (Length(FVolume) > 0) then
				begin { Label Disk }
				StrPLCopy(PChar(@Temp),FVolume,Length(FVolume));
				if FMakeBootable <> True then
					Offset := 0
				else
					Offset := 32*3; { Leave Room for System }
				Move(Temp,Buffer^[Offset+1],Length(FVolume));
				Buffer^[Offset+12] := $28;  { File Attribute }
				FileTime := SysUtils.DateTimeToFileDate(SysUtils.Now);
				Move(FileTime,Buffer^[Offset+23],Sizeof(LongInt));
				end; { Label Disk }
			Result := WriteLogicalSector(Drive,LogicalSector,Buffer);
			if Result <> 0 then
				exit;
			end;
		end;
end;

function TFormat.WriteDisk(Drive : Byte; Head : Byte; Track : Byte; Sector : Byte;
NumSecs : Byte; Buffer : Pointer) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $3;
		Regs.al := NumSecs;
		Regs.cl := Sector;
		Regs.ch := Track;
		Regs.dh := Head;
		Regs.es := Seg(Buffer^);
		Regs.bx := Ofs(Buffer^);
		Regs.dl := Drive;
		if Regs.dl <= 1 then { Never can be to safe }
			intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			break;
		ResetDrive(Drive);
		end;
end;

function TFormat.WriteLogicalSector(Drive : Byte; LogicalSector : Integer; Buffer : Pointer) : Integer;
var
Head : Byte;
Track : Byte;
Sector : Byte;
begin
	with DiskInfo[Ord(FActualMediaType)] do
		begin
		Head  := (LogicalSector div SectorsPerTrack) mod Heads;
		Track := (LogicalSector div SectorsPerTrack) div Heads;
		Sector := (LogicalSector mod SectorsPerTrack) + 1;
		Result := WriteDisk(Drive,Head,Track,Sector,1,Buffer);
		end;
end;

function TFormat.ReadDisk(Drive : Byte; Head : Byte; Track : Byte; Sector : Byte;
NumSecs : Byte; Buffer : Pointer) : Integer;
var
Retry : Integer;
Regs : TRegisters;
begin
	for Retry := 1 to Retries do
		begin
		FillChar(Regs,Sizeof(Regs),0);
		Regs.ah := $2;
		Regs.al := NumSecs;
		Regs.cl := Sector;
		Regs.ch := Track;
		Regs.dh := Head;
		Regs.es := Seg(Buffer^);
		Regs.bx := Ofs(Buffer^);
		Regs.dl := Drive;
		intr($13,Regs);
		Result := Regs.ah;
		if Result = 0 then
			break;
		ResetDrive(Drive);
		end;
end;


function TFormat.SerialNumber : LongInt;
var
Year,Month,Day : Word;
Hour,Min,Sec,MSec : Word;
SerialNumber : array[1..2] of Integer;
begin
	SysUtils.DecodeDate(SysUtils.Date,Year,Month,Day);
	SysUtils.DecodeTime(SysUtils.Time,Hour,Min,Sec,MSec);
	SerialNumber[1] := Year + ((Hour shl 8) or Min);
	SerialNumber[2] := ((Month shl 8) or Day) + ((Sec shl 8)or MSec);
	Result := LongInt(SerialNumber);
end;

function TFormat.MakeBootDisk(Drive : Char) : Integer;
var
FileName : String[13];
Source : String;
Dest : String;
begin
	Result := -1;

	{ Lets Remove any system file that may exist }
	Dest := Drive + ':' + _IOSYS;
	FileSetAttr(Dest,0);
	DeleteFile(Dest);
	Dest := Drive + ':' + _MSDOS;
	FileSetAttr(Dest,0);
	DeleteFile(Dest);
	Dest := Drive + ':' + _IBMBIO;
	FileSetAttr(Dest,0);
	DeleteFile(Dest);
	Dest := Drive + ':' + _IBMDOS;
	FileSetAttr(Dest,0);
	DeleteFile(Dest); 

	if IsMSDOS = True then
		begin { MSDOS }
		FileName := _IOSYS;
		Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + FileName;
		Dest := Drive + ':' + FileName;
		Result := CopyFile(Source,Dest);

		if Result = 0 then
			begin
			FileName := _MSDOS;
			Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + FileName;
			Dest := Drive + ':' + FileName;
			Result := CopyFile(Source,Dest);
			end; { MSDOS}
		end
	else
		begin { IBMDOS }
		FileName := _IBMBIO;
		Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + FileName;
		Dest := Drive + ':' + FileName;
		Result := CopyFile(Source,Dest);

		if Result = 0 then
			begin
			FileName := _IBMDOS;
			Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + FileName;
			Dest := Drive + ':' + FileName;
			Result := CopyFile(Source,Dest);
			end;
		end; { IBMDOS }

	if Result = 0 then
		begin
		Source := StrPas(GetEnvVar('COMSPEC'));
		if Source[2] <> ':' then
			Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + StrPas(GetEnvVar('COMSPEC'));
		Dest := Drive + ':' + '\COMMAND.COM';
		Result := CopyFile(Source,Dest);
		end;
end;

function TFormat.CopyFile(Source : String; Dest : String) : Integer;
var
Buffer : Pointer;
FPSource : Integer;
FPDest : Integer;
Attr : Integer;
DateAndTime : LongInt;
BufSize : LongInt;
AmtRead, AmtWrite : LongInt;
begin
	FPSource := FileOpen(Source,fmOpenRead);
	if FPSource < 0 then
		begin
		Result := FPSource;
		exit;
		end;

	FileSetAttr(Dest,0);
	FPDest := FileCreate(Dest);
	if FPDest < 0 then
		begin;
		Result := FPDest;
		exit;
		end;

	BufSize := 40960;
	GetMem(Buffer,BufSize);

	repeat
		AmtRead := FileRead(FPSource,Buffer^,BufSize);
		AmtWrite := FileWrite(FPDest,Buffer^,AmtRead);
	until (AmtRead = 0) or (AmtWrite = 0);

	if AmtRead <> AmtWrite then
		Result := 1
	else
		begin
		Result := 0;

		{ Get and set file time }
		DateAndTime := FileGetDate(FPSource);
		FileSetDate(FPDest,DateAndTime);
		end;

	FreeMem(Buffer,BufSize);
	FileClose(FPSource);
	FileClose(FPDest);

	if Result = 0 then
		begin
		{ Get and set file attributes }
		Attr := FileGetAttr(Source);
		FileSetAttr(Dest,Attr);
		end;
end;

function TFormat.IsMSDOS : Boolean;
var SearchRec : SysUtils.TSearchRec;
Source : String;
begin
	Source := Char(Integer('C')+Ord(FSystemLocation)) + ':' + _IBMBIO;
	if SysUtils.FindFirst(Source,faAnyFile,SearchRec) = 0 then
		Result := False
	else
		Result := True;
	FindClose(SearchRec);
end;

function TFormat.CheckFormatConditions(Drive : Byte; var TempFormatCondition : TFormatCondition) : Integer;
var
Buffer : MaxArrayPtr;
Error : Integer;
CurrentDiskInfo : DiskInformation;
Matched : Byte;
begin
	Result := 0;
	if (TempFormatCondition = WhenNeeded) or ((TempFormatCondition = Never) and (ClearDisk = True)) then
		begin
		{ Hope this is big enough }
		GetMem(Buffer,2048);

		{ Read in boot sector }
		Error := ReadDisk(Drive,0,0,1,1,Buffer);
		Move(Buffer^[12],CurrentDiskInfo,Sizeof(CurrentDiskInfo));

		FreeMem(Buffer,2048);

		Matched := 1;
		while (Error = 0) and (Matched <= 4) do
			begin
			if CompareMemory(MaxArrayPtr(@CurrentDiskInfo),MaxArrayPtr(@DiskInfo[Matched]),Sizeof(DiskInformation)) = 0 then
				break;
			Matched := Matched + 1;
			end;

		if TempFormatCondition = Never then
			begin { ClearDisk }
			if FMediaType = BestFit then
				begin { BestFit }
				if (Error <> 0) or (Matched > 4) then
					begin
					{ Error or Boot is not regognized }
					Result := 14;
					exit;
					end;
				FActualMediaType := TMediaType(Matched);
				end { BestFit }
			else
				begin { Not BestFit }
				if (Error <> 0) or (FActualMediaType <> TMediaType(Matched)) then
					begin
					{ Error or Boot does not match current MediaType }
					Result := 14;
					exit
					end;
				end;
			end { ClearDisk }
		else
			begin { WhenNeeded }
			if FMediaType = BestFit then
				begin { BestFit }
				if (Error <> 0) or (Matched > 4) then
					{ Error or Boot is not regognized }
					TempFormatCondition := Always
				else
					{ Use MediaType on disk }
					FActualMediaType := TMediaType(Matched);
				end { BestFit }
			else
				begin { Not BestFit }
				if (Error <> 0) or (FActualMediaType <> TMediaType(Matched)) then
					{ Error or Boot does not match current MediaType }
				TempFormatCondition := Always;
				end;
			end;
		end;
end;

function TFormat.CompareMemory(Value1, Value2 : MaxArrayPtr; Size : Integer) : Integer;
var Pos : Integer;
begin
	Result := 0;
	for Pos := 1 to Size do
		begin
		if Value1^[Pos] > Value2^[Pos] then 
			begin
			Result := 1;
			break;
			end;
		if Value1^[Pos] < Value2^[Pos] then
			begin
			Result := -1;
			break;
			end;
		end;
end;

function TFormat.GetErrNoMessage(ErrNo : Integer) : String;
begin
	case ErrNo of
		$00 : Result := 'OK';
		$01 : Result := 'Invalid command';
		$02 : Result := 'Address not found';
		$03 : Result := 'Disk write protected';
		$04 : Result := 'Sector not found';
		$06 : Result := 'Floppy disk removed';
		$08 : Result := 'DMA overflow';
		$09 : Result := 'DMA transmission beyond segment boarder';
		$0C : Result := 'Media type not found';
		$10 : Result := 'Read Error';
		$20 : Result := 'Controller failed';
		$40 : Result := 'Seek failed';
		$80 : Result := 'Disk timed out';
	else
		Result := 'Unknown error '+ IntToStr(ErrNo);
   end;
end;

function TFormat.GetErrorMessage(Error : Integer) : String;
begin
	case Error of
		0 : Result := 'OK';
		1 : Result := 'Reset failed';
		2 : Result := 'Unable to get drive type';
		3 : Result := 'No such drive';
		4 : Result := 'Unknown drive type';
		5 : Result := 'Incompatible media types';
		6 : Result := 'Unable to set disk type';
		7 : Result := 'Unable to set media type';
		8 : Result := 'Error formatting track';
		9 : Result := 'Error verifying track';
		10: Result := 'Error writing boot sector';
		11: Result := 'Error writing fat';
		12: Result := 'Error writing root';
		13: Result := 'Error creating system';
		14: Result := 'Unable to clear disk';
		98: Result := 'Process already running';
		99: Result := 'Canceled by user';
	else
		Result := 'Unknown error '+ IntToStr(Error);
   	end;
end;

end.
